"""Runner output processor default"""

import sys
import typing as t

import inquirer  # type: ignore

from .base import BaseDisplay
from .color import Color
from ..actions.base import ActionBase, ActionStatus
from ..actions.types import Stderr
from ..exceptions import InteractionError
from ..workflow import Workflow

__all__ = [
    "PrologueDisplay",
    "HeaderDisplay",
    "PrefixDisplay",
    "DefaultDisplay",
    "KNOWN_DISPLAYS",
]


class PrologueDisplay(BaseDisplay):
    """Default display base"""

    NAME: str

    STATUS_TO_COLOR_WRAPPER_MAP: t.Dict[ActionStatus, t.Callable[[str], str]] = {
        ActionStatus.SKIPPED: Color.gray,
        ActionStatus.PENDING: Color.gray,
        ActionStatus.FAILURE: Color.red,
        ActionStatus.WARNING: Color.yellow,
        ActionStatus.RUNNING: lambda x: x,
        ActionStatus.SUCCESS: Color.green,
        ActionStatus.OMITTED: Color.gray,
    }

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        self._last_displayed_name: t.Optional[str] = None

    def _make_prologue(self, source: ActionBase, mark: str) -> str:
        raise NotImplementedError

    def on_action_message(self, source: ActionBase, message: str) -> None:
        is_stderr: bool = isinstance(message, Stderr)
        for line in message.splitlines() if message else [message]:
            line_prefix: str = self._make_prologue(source=source, mark="*" if is_stderr else " ")
            self.display(f"{line_prefix}{Color.yellow(line) if is_stderr else line}")

    def on_action_error(self, source: ActionBase, message: str) -> None:
        line_prefix: str = self._make_prologue(source=source, mark="!")
        for line in message.splitlines():
            super().on_action_error(
                source=source,
                message=f"{line_prefix}{Color.red(line)}",
            )

    def _display_status_banner(self) -> None:
        """Show a text banner with the status info"""
        raise NotImplementedError

    def on_runner_finish(self) -> None:
        self._display_status_banner()

    def on_plan_interaction(self, workflow: Workflow) -> None:
        displayed_action_names_with_descriptions: t.List[str] = []
        default_selected_action_names_with_descriptions: t.List[str] = []
        full_description_to_names_map: t.Dict[str, str] = {}
        for _, action in workflow.iter_actions_by_tier():
            if action.selectable:
                action_name_with_description: str = action.name
                if action.description is not None:
                    action_name_with_description = f"{action_name_with_description}: {action.description}"
                displayed_action_names_with_descriptions.append(action_name_with_description)
                default_selected_action_names_with_descriptions.append(action_name_with_description)
                if action_name_with_description in full_description_to_names_map:
                    raise InteractionError(f"Action full descriptions collision: {action_name_with_description!r}")
                full_description_to_names_map[action_name_with_description] = action.name
        if not displayed_action_names_with_descriptions:
            raise InteractionError("No selectable actions found")
        selected_action_names: t.List[str] = self._run_dialog(
            choices=displayed_action_names_with_descriptions,
            default=default_selected_action_names_with_descriptions,
        )
        self.logger.warning(f"Interactively selected actions: {selected_action_names}")
        for action_name, action in workflow.items():
            if action_name in displayed_action_names_with_descriptions and action_name not in selected_action_names:
                action.disable()

    @classmethod
    def _run_dialog(cls, choices: t.List[str], default: t.List[str]) -> t.List[str]:  # pragma: no cover
        if not sys.stdin.isatty():
            raise InteractionError
        answers: t.Dict[str, t.List[str]] = inquirer.prompt(
            questions=[
                inquirer.Checkbox(
                    name="actions",
                    message="Select actions (SPACE to check, RETURN to proceed)",
                    choices=choices,
                    default=default,
                    carousel=True,
                )
            ],
            raise_keyboard_interrupt=True,
        )
        selected_action_names: t.List[str] = answers["actions"]
        return selected_action_names


class PrefixDisplay(PrologueDisplay):
    """Adds prefixes to output chunks"""

    NAME = "prefixes"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        self._action_names_max_len = max(map(len, self._workflow))

    def _make_prologue(self, source: ActionBase, mark: str) -> str:
        """Construct prefix based on previous emitter action name"""
        justification_len: int = self._action_names_max_len + 2  # "2" here stands for square brackets
        formatted_name: str = (
            f"[{source.name}]".ljust(justification_len)
            if self._last_displayed_name != source.name
            else " " * justification_len
        )
        self._last_displayed_name = source.name
        return Color.gray(f"{formatted_name} {mark}| ")

    def _display_status_banner(self) -> None:
        justification_len: int = self._action_names_max_len + 9  # "9" here stands for (e.g.) "SUCCESS: "
        self.display(Color.gray("=" * justification_len))
        for _, action in self._workflow.iter_actions_by_tier():
            color_wrapper: t.Callable[[str], str] = self.STATUS_TO_COLOR_WRAPPER_MAP[action.status]
            self.display(f"{color_wrapper(action.status.value)}: {action.name}")


class HeaderDisplay(PrologueDisplay):
    """Adds headers to output chunks"""

    NAME = "headers"
    _STATUS_TO_MARK_SYMBOL_MAP: t.Dict[ActionStatus, str] = {
        ActionStatus.SKIPPED: "◯",
        ActionStatus.PENDING: "◯",
        ActionStatus.FAILURE: "✗",
        ActionStatus.WARNING: "✓",
        ActionStatus.RUNNING: "◯",
        ActionStatus.SUCCESS: "✓",
        ActionStatus.OMITTED: "◯",
    }

    def _close_block_if_necessary(self) -> None:
        if self._last_displayed_name is not None:
            self.display(Color.gray(" ╵"))

    def _make_prologue(self, source: ActionBase, mark: str) -> str:
        """Construct header based on previous emitter action name"""
        if self._last_displayed_name != source.name:
            self._close_block_if_necessary()
            self.display(Color.gray(f" ┌─[{source.name}]"))
            self._last_displayed_name = source.name
        return Color.gray(f"{mark}│ ")

    def _display_status_banner(self) -> None:
        self._close_block_if_necessary()
        for _, action in self._workflow.iter_actions_by_tier():
            color_wrapper: t.Callable[[str], str] = self.STATUS_TO_COLOR_WRAPPER_MAP[action.status]
            mark_symbol: str = self._STATUS_TO_MARK_SYMBOL_MAP[action.status]
            state_string = f" {mark_symbol} {action.status.value}: {action.name}"
            self.display(color_wrapper(state_string))


DefaultDisplay = PrefixDisplay
KNOWN_DISPLAYS: t.Dict[str, t.Type[BaseDisplay]] = {
    HeaderDisplay.NAME: HeaderDisplay,
    PrefixDisplay.NAME: PrefixDisplay,
}



"""
Runner has too many dependencies,
thus placed to a separate module.
"""

import asyncio
import functools
import io
import sys
import typing as t
from enum import Enum
from pathlib import Path

import classlogging
import dacite

from . import types
from .actions.base import ActionBase, ArgsBase, ActionStatus
from .config.constants import C
from .display.base import BaseDisplay
from .exceptions import SourceError, ExecutionFailed, ActionRenderError, ActionRunError
from .loader.helpers import get_default_loader_class_for_source
from .rendering import Templar
from .tools.concealment import represent_object_type
from .workflow import Workflow

__all__ = [
    "Runner",
]

IOType = io.TextIOBase
logger = classlogging.get_module_logger()


class Runner(classlogging.LoggerMixin):
    """Main entry object"""

    def __init__(
        self,
        source: t.Union[str, Path, IOType, None] = None,
        display: t.Optional[types.DisplayType] = None,
    ) -> None:
        self._workflow_source: t.Union[Path, IOType] = self._detect_workflow_source(explicit_source=source)
        self._explicit_display: t.Optional[types.DisplayType] = display
        self._started: bool = False
        self._outcomes: t.Dict[str, t.Dict[str, t.Any]] = {}
        self._execution_failed: bool = False

    @functools.cached_property
    def loader(self) -> types.LoaderType:
        """Workflow loader"""
        loader_class: types.LoaderClassType
        if C.WORKFLOW_LOADER_CLASS is not None:
            loader_class = C.WORKFLOW_LOADER_CLASS
        else:
            loader_class = get_default_loader_class_for_source(self._workflow_source)
        self.logger.debug(f"Using workflow loader class: {loader_class}")
        return loader_class()

    @functools.cached_property
    def workflow(self) -> Workflow:
        """Calculated workflow"""
        return (
            self.loader.loads(self._workflow_source.read())
            if isinstance(self._workflow_source, io.TextIOBase)
            else self.loader.load(self._workflow_source)
        )

    @functools.cached_property
    def display(self) -> types.DisplayType:
        """Attached display"""
        if self._explicit_display is not None:
            self.logger.debug(f"Using explicit display: {self._explicit_display}")
            return self._explicit_display
        display_class: types.DisplayClassType = C.DISPLAY_CLASS
        self.logger.debug(f"Using display class: {display_class}")
        return display_class(workflow=self.workflow)

    @functools.cached_property
    def strategy(self) -> types.StrategyType:
        """Strategy iterator"""
        if self.loader.strategy_class is not None:
            strategy_class: types.StrategyClassType = self.loader.strategy_class
            self.logger.debug(f"Using strategy class from the loaded workflow: {strategy_class}")
        else:
            strategy_class = C.STRATEGY_CLASS
            self.logger.debug(f"Using globally-set strategy class: {strategy_class}")
        return strategy_class(workflow=self.workflow)

    @classmethod
    def _detect_workflow_source(cls, explicit_source: t.Union[str, Path, IOType, None] = None) -> t.Union[Path, IOType]:
        if explicit_source is not None:
            if isinstance(explicit_source, IOType):
                return explicit_source
            return Path(explicit_source)
        if C.ACTIONS_SOURCE_FILE is not None:
            source_file: Path = C.ACTIONS_SOURCE_FILE
            if str(source_file) == "-":
                cls.logger.info("Using stdin as workflow source")
                return t.cast(IOType, sys.stdin)
            if not source_file.exists():
                raise SourceError(f"Given workflow file does not exist: {source_file}")
            cls.logger.info(f"Using given workflow file: {source_file}")
            return source_file
        scan_path: Path = C.CONTEXT_DIRECTORY
        cls.logger.debug(f"Looking for workflow files at {str(scan_path)!r}")
        located_source_file: t.Optional[Path] = None
        for candidate_file_name in (
            "grana.yml",
            "grana.yaml",
        ):  # type: str
            if (maybe_source_file := scan_path / candidate_file_name).exists():
                cls.logger.info(f"Detected the workflow source: {str(maybe_source_file)!r}")
                if located_source_file is not None:
                    raise SourceError(f"Multiple workflow sources detected in {scan_path}")
                located_source_file = maybe_source_file
        if located_source_file is None:
            raise SourceError(f"No workflow source detected in {scan_path}")
        return located_source_file

    async def run_async(self) -> None:
        """Primary coroutine for all further processing"""
        # Build workflow and display
        workflow: Workflow = self.workflow
        display: BaseDisplay = self.display
        # Check requirements
        self.loader.check_requirements()
        try:
            display.on_runner_start()
        except Exception:
            self.logger.exception(f"`on_runner_start` callback failed for {display}")
        if C.INTERACTIVE_MODE:
            display.on_plan_interaction(workflow=workflow)
        await self._run_all_actions()
        try:
            display.on_runner_finish()
        except Exception:
            self.logger.exception(f"`on_runner_finish` callback failed for {display}")
        if self._execution_failed:
            raise ExecutionFailed

    async def _run_all_actions(self) -> None:
        if self._started:
            raise RuntimeError("Runner has been started more than one time")
        self._started = True
        action_runners: t.Dict[ActionBase, asyncio.Task] = {}
        # Prefill outcomes map
        for action_name in self.workflow:
            self._outcomes[action_name] = {}
        async for action in self.strategy:  # type: ActionBase
            # Finalize all actions that have been done already
            for maybe_finished_action, corresponding_runner_task in list(action_runners.items()):
                if maybe_finished_action.done():
                    self.logger.trace(f"Finalizing done action {maybe_finished_action.name!r} runner")
                    await corresponding_runner_task
                    action_runners.pop(maybe_finished_action)
            self.logger.trace(f"Allocating action runner for {action.name!r}")
            action_runners[action] = asyncio.create_task(self._run_action(action=action))

        # Finalize running actions
        for task in action_runners.values():
            await task

    @classmethod
    async def _dispatch_action_messages_to_display(cls, action: ActionBase, display: BaseDisplay) -> None:
        try:
            async for message in action.read_messages():
                display.on_action_message(source=action, message=message)
        except Exception:
            cls.logger.exception(f"`on_action_message` failed for {action.name!r}")

    async def _run_action(self, action: ActionBase) -> None:
        if not action.enabled:
            action._internal_omit()  # pylint: disable=protected-access
            return None
        message: str
        try:
            self._render_action(action)
        except Exception as e:
            details: str = str(e) if isinstance(e, ActionRenderError) else repr(e)
            message = f"Action {action.name!r} rendering failed: {details}"
            self.display.on_action_error(source=action, message=message)
            self.logger.warning(message, exc_info=not isinstance(e, ActionRenderError))
            action._internal_fail(e)  # pylint: disable=protected-access
            self._execution_failed = True
            return
        self.logger.trace(f"Calling `on_action_start` for {action.name!r}")
        try:
            self.display.on_action_start(action)
        except Exception:
            self.logger.exception(f"`on_action_start` callback failed on {action.name!r} for {self.display}")
        self.logger.trace(f"Allocating action dispatcher for {action.name!r}")
        action_messages_reader_task: asyncio.Task = asyncio.create_task(
            self._dispatch_action_messages_to_display(
                action=action,
                display=self.display,
            )
        )
        try:
            await action
        except Exception as e:
            try:
                self.display.on_action_error(
                    source=action,
                    message=str(e) if isinstance(e, ActionRunError) else f"Action {action.name!r} run exception: {e!r}",
                )
            except Exception:
                self.logger.exception(f"`on_action_error` failed for {action.name!r}")
            if action.status == ActionStatus.WARNING:
                self.logger.warning(f"Action {action.name!r} finished with warning status")
            else:
                self.logger.warning(f"Action {action.name!r} execution failed: {e!r}")
                self._execution_failed = True
            self.logger.debug("Action failure traceback", exc_info=True)
        finally:
            self._outcomes[action.name].update(action.get_outcomes())
            await action_messages_reader_task
            self.logger.trace(f"Calling `on_action_finish` for {action.name!r}")
            try:
                self.display.on_action_finish(action)
            except Exception:
                self.logger.exception(f"`on_action_finish` callback failed on {action.name!r} for {self.display}")

    def run_sync(self):
        """Wrap async run into an event loop"""
        asyncio.run(self.run_async())

    def _render_action(self, action: ActionBase) -> None:
        """Prepare action to execution by rendering its template fields"""
        templar: Templar = Templar(
            outcomes_map=self._outcomes,
            action_states={name: self.workflow[name].status.value for name in self.workflow},
            context_map=self.workflow.context,
        )

        rendered_args_dict: dict = templar.recursive_render(self.loader.get_original_args_dict_for_action(action))
        try:
            parsed_args: ArgsBase = dacite.from_dict(
                data_class=type(action.args),
                data=rendered_args_dict,
                config=dacite.Config(
                    strict=True,
                    cast=[Enum, Path],
                ),
            )
        except dacite.WrongTypeError as e:
            raise ActionRenderError(
                f"Unrecognized {e.field_path!r} content type: {represent_object_type(e.value)}"
                f" (expected {e.field_type!r})"
            ) from None
        action.args = parsed_args



"""Types collection"""

import dataclasses
import typing as t

OutcomeStorageType = t.Dict[str, str]

__all__ = [
    "OutcomeStorageType",
    "Stderr",
    "Import",
    "ObjectTemplate",
    "qualify_string_as_potentially_renderable",
]


class Stderr(str):
    """Strings related to standard error stream"""


@dataclasses.dataclass
class Import:
    """Import clause"""

    path: str


@dataclasses.dataclass
class ObjectTemplate:
    """Complex object expression to be rendered later"""

    expression: str


def qualify_string_as_potentially_renderable(data: str) -> bool:
    """Check that a string should be templated later"""
    return "@{" in data



"""Templar containers for rendering."""

import typing as t

import lazy_object_proxy  # type: ignore

from ..exceptions import ActionRenderError

__all__ = [
    "AttrDict",
    "LooseDict",
    "StrictOutcomeDict",
    "ActionContainingDict",
    "ContextDict",
    "LazyProxy",
]

RenderHookType = t.Callable[[str], str]


class ItemAttributeAccessorMixin:
    """Anything, that can be accessed fie __getitem__, is available also as an attribute"""

    def __getattr__(self, item: str):
        return self.__getitem__(item)


class AttrDict(dict, ItemAttributeAccessorMixin):
    """Basic dictionary that allows attribute read access to its keys"""


class LooseDict(AttrDict):
    """A dictionary that allows attribute read access to its keys with a default empty value fallback"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError:
            return ""


class StrictOutcomeDict(AttrDict):
    """A dictionary that allows attribute read access to its keys with a default value fallback"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise ActionRenderError(f"Outcome key {e} not found") from e


class ActionContainingDict(AttrDict):
    """Anything with action names as keys"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise ActionRenderError(f"Action not found: {e}") from e


class ContextDict(AttrDict):
    """Context keys representation"""

    def __getitem__(self, item: str):
        # Context keys can refer to anything else, thus we keep resolving until the template is stable
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise ActionRenderError(f"Context key not found: {e}") from e


class LazyProxy(lazy_object_proxy.Proxy):
    """Lazy proxy that `repr`s like its wrapped object"""

    def __repr__(self, __getattr__=object.__getattribute__) -> str:
        return repr(self.__wrapped__)
